import json
import logging
import re
import subprocess
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any, AsyncGenerator, Literal, cast, overload
from uuid import UUID

import httpx
from moxn.types.base import BranchHeadResponse, RenderableModel, VersionRef
from moxn.types.content import Provider
from moxn.types.requests import MessageData, PromptCreateRequest, TaskCreateRequest
from moxn.types.responses import DatamodelCodegenResponse
from moxn.types.telemetry import (
    MoxnTraceCarrier,
    Span,
    SpanContext,
    TelemetryLogRequest,
    TelemetryLogResponse,
    TraceContext,
)
from moxn.types.type_aliases.anthropic import AnthropicMessage
from moxn.types.type_aliases.google import GoogleGenerateContentResponse
from moxn.types.type_aliases.openai_chat import OpenAIChatCompletion

from moxn.base_models.blocks.signed import (
    SignedURLContent,
    SignedURLImageContent,
    SignedURLPDFContent,
)
from moxn.content.client import ContentClient
from moxn.models.prompt import PromptSession, PromptTemplate
from moxn.models.response import LLMEvent
from moxn.models.task import Task
from moxn.settings import MoxnSettings
from moxn.storage.storage import InMemoryStorage
from moxn.telemetry.client import TelemetryClient

logger = logging.getLogger(__name__)


class MoxnClient:
    """
    Moxn API client for interacting with the Moxn platform.

    Example:
        ```python
        # Automatically uses MOXN_API_KEY from environment
        async with MoxnClient() as client:
            session = await client.create_prompt_session(
                prompt_id=f"{uuid}",
                branch_name="main"  # or commit_id="{commit_id}"
            )
        ```
    """

    def __init__(self) -> None:
        self.settings = MoxnSettings()  # type: ignore
        self._client: httpx.AsyncClient | None = None
        self.storage = InMemoryStorage()
        self._context_depth = 0  # Track nested context usage
        self._registered_file_paths: set[str] = set()  # Track registered signed content

        self.telemetry_client = TelemetryClient.from_settings(self.settings)
        self.content_client = ContentClient.from_settings(self.settings)

    @property
    def client(self) -> httpx.AsyncClient:
        """Get the HTTP client, creating it if necessary."""
        if self._client is None:
            self._client = self._create_client()
        return self._client

    def _create_client(self) -> httpx.AsyncClient:
        """Creates an authenticated httpx client."""
        return httpx.AsyncClient(
            base_url=str(self.settings.base_api_route),
            timeout=self.settings.timeout,
            headers=self.get_headers(),
        )

    def get_headers(self, _: bool = True) -> dict:
        return _get_headers(self.settings, _)

    async def __aenter__(self) -> "MoxnClient":
        self._context_depth += 1
        if self._client is None:
            self._client = self._create_client()
        await self.telemetry_client.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self._context_depth -= 1
        if self._context_depth == 0:
            await self.telemetry_client.stop()
            # Polling removed - no cleanup needed
            await self.close()

    async def close(self) -> None:
        """Close the underlying HTTP client."""
        if self._client is not None:
            await self._client.aclose()
            self._client = None

    def get_telemetry_diagnostics(self) -> dict:
        """Get diagnostic information from telemetry system."""
        dispatcher_diag = {}
        if (
            hasattr(self.telemetry_client, "_dispatcher")
            and self.telemetry_client._dispatcher
        ):
            dispatcher_diag = self.telemetry_client._dispatcher.get_diagnostics()

        return {
            "dispatcher": dispatcher_diag,
            "telemetry_started": getattr(self.telemetry_client, "_started", False),
        }

    async def verify_access(self) -> dict:
        """Verify API access and return authenticated identity information.

        This method calls the gateway's /debug/whoami endpoint to verify that:
        - The API key is valid
        - Authentication is working correctly
        - The gateway can communicate with the backend

        Returns:
            dict: Authentication information including:
                - authenticated: bool
                - tenant_id: str
                - message: str

        Raises:
            httpx.HTTPStatusError: If authentication fails
        """
        response = await self.get("/debug/whoami")
        response.raise_for_status()
        return response.json()

    def _sanitize_module_name(self, filename: str) -> str:
        """
        Sanitize filename to ensure it's a valid Python module name.

        Args:
            filename: Original filename (e.g., "sdk-test_models.py")

        Returns:
            Sanitized filename (e.g., "sdk_test_models.py")
        """
        # Remove .py extension if present
        name = filename[:-3] if filename.endswith(".py") else filename

        # Replace hyphens with underscores
        name = name.replace("-", "_")

        # Replace spaces with underscores
        name = name.replace(" ", "_")

        # Remove any characters that aren't alphanumeric or underscore
        name = re.sub(r"[^a-zA-Z0-9_]", "", name)

        # Ensure it doesn't start with a digit
        if name and name[0].isdigit():
            name = f"_{name}"

        # Add .py extension back
        return f"{name}.py" if filename.endswith(".py") else name

    def _format_with_ruff(self, file_path: Path) -> None:
        """
        Attempt to format a generated Python file with ruff.

        This is a best-effort operation - if ruff is not installed or fails,
        the file is left as-is and a warning is logged.

        Args:
            file_path: Path to the Python file to format
        """
        try:
            logger.info(f"Formatting {file_path} with ruff...")
            result = subprocess.run(
                ["ruff", "format", str(file_path)],
                capture_output=True,
                text=True,
                timeout=30,
            )
            if result.returncode == 0:
                logger.info(f"Successfully formatted {file_path} with ruff")
            else:
                logger.warning(
                    f"ruff format returned non-zero exit code: {result.returncode}. "
                    f"stderr: {result.stderr}"
                )
        except FileNotFoundError:
            logger.warning(
                "ruff is not installed or not in PATH. "
                "Skipping code formatting. Install with: pip install ruff"
            )
        except subprocess.TimeoutExpired:
            logger.warning(f"ruff format timed out for {file_path}")
        except Exception as e:
            logger.warning(f"Failed to format {file_path} with ruff: {e}")

    async def get(self, path: str, **kwargs) -> httpx.Response:
        """Perform a GET request."""
        return await self.client.get(path, **kwargs)

    async def post(self, path: str, **kwargs) -> httpx.Response:
        """Perform a POST request."""
        return await self.client.post(path, **kwargs)

    async def put(self, path: str, **kwargs) -> httpx.Response:
        """Perform a PUT request."""
        return await self.client.put(path, **kwargs)

    async def delete(self, path: str, **kwargs) -> httpx.Response:
        """Perform a DELETE request."""
        return await self.client.delete(path, **kwargs)

    def _unwrap_gateway_response(self, response_data: dict) -> dict:
        """
        Unwrap the gateway's standardized APIResponse format.

        Expected format: {"success": bool, "data": {...}, "error": null}
        Returns the "data" field if successful, raises ValueError if not.
        """
        if not response_data.get("success"):
            error_msg = response_data.get("error", "Unknown error")
            raise ValueError(f"Gateway returned error: {error_msg}")

        return response_data["data"]

    async def get_branch_head(
        self, task_id: str, branch_name: str = "main"
    ) -> BranchHeadResponse:
        """
        Get the head commit for a branch.

        Use this during development to get the latest commit on a branch.
        In production, use explicit commit IDs.
        """
        params = {"task_id": task_id}  # Fixed: use snake_case for FastAPI endpoint
        response = await self.get(f"/branches/{branch_name}/head", params=params)
        response.raise_for_status()

        response_data = self._unwrap_gateway_response(response.json())
        return BranchHeadResponse.model_validate(response_data)

    async def _register_signed_content_in_messages(self, messages: list) -> None:
        """Register all signed content blocks for automatic URL refresh.

        Recursively scans message blocks and registers any SignedURL content
        with the content client's refresh engine. Uses file_path-based deduplication
        to prevent duplicate heap entries for the same storage file.

        Args:
            messages: List of messages to scan for signed content
        """

        for message in messages:
            for block_group in message.blocks:
                for block in block_group:
                    # Check for base SignedURLContent and all subclasses
                    if isinstance(
                        block,
                        (SignedURLContent, SignedURLImageContent, SignedURLPDFContent),
                    ):
                        # Only register if we haven't seen this file_path yet
                        if block.file_path not in self._registered_file_paths:
                            await self.content_client.register_content(block)
                            self._registered_file_paths.add(block.file_path)

    async def fetch_task(self, task_id: str, version: VersionRef) -> Task:
        """
        Fetch a task from the gateway API. Always fetches directly from API.
        """
        params = {}
        if version.branch_name:
            params["branch_name"] = version.branch_name
        else:
            assert version.commit_id is not None
            params["commit_id"] = version.commit_id

        response = await self.get(f"/tasks/{task_id}", params=params)
        response.raise_for_status()

        task_data = self._unwrap_gateway_response(response.json())
        task = Task.model_validate(task_data)

        # Register signed content for auto-refresh from all prompts in task
        for prompt in task.prompts:
            await self._register_signed_content_in_messages(prompt.messages)

        return task

    async def fetch_prompt(self, prompt_id: str, version: VersionRef) -> PromptTemplate:
        """
        Fetch a prompt from the gateway API. Always fetches directly from API.
        """
        params = {}
        if version.branch_name:
            params["branch_name"] = version.branch_name
        else:
            assert version.commit_id is not None
            params["commit_id"] = version.commit_id

        response = await self.get(f"/prompts/{prompt_id}", params=params)
        response.raise_for_status()

        prompt_data = self._unwrap_gateway_response(response.json())
        prompt = PromptTemplate.model_validate(prompt_data)

        # Register signed content for auto-refresh
        await self._register_signed_content_in_messages(prompt.messages)

        return prompt

    async def get_prompt(
        self,
        prompt_id: UUID | str,
        branch_name: str | None = None,
        commit_id: str | None = None,
    ) -> PromptTemplate:
        """
        Get a prompt with caching. Commits are cached (immutable), branches always fetch latest.

        Args:
            prompt_id: The prompt ID
            branch_name: Branch name (mutually exclusive with commit_id)
            commit_id: Commit ID (mutually exclusive with branch_name)
        """
        if not isinstance(prompt_id, str):
            prompt_id = str(prompt_id)

        version = VersionRef(branch_name=branch_name, commit_id=commit_id)

        # Only cache commits (immutable), always fetch branches (latest)
        if version.commit_id:
            try:
                return await self.storage.get_prompt_by_commit(
                    prompt_id, version.commit_id
                )
            except KeyError:
                pass

            prompt = await self.fetch_prompt(prompt_id, version)
            await self.storage.store_prompt(prompt, commit_id=version.commit_id)
            return prompt
        else:
            # Always fetch branches to get latest
            return await self.fetch_prompt(prompt_id, version)

    async def get_task(
        self,
        task_id: str,
        branch_name: str | None = None,
        commit_id: str | None = None,
    ) -> Task:
        """
        Get a task with caching. Commits are cached (immutable), branches always fetch latest.

        Args:
            task_id: The task ID
            branch_name: Branch name (mutually exclusive with commit_id)
            commit_id: Commit ID (mutually exclusive with branch_name)
        """

        version = VersionRef(branch_name=branch_name, commit_id=commit_id)

        # Only cache commits (immutable), always fetch branches (latest)
        if version.commit_id:
            try:
                return await self.storage.get_task_by_commit(task_id, version.commit_id)
            except KeyError:
                pass

            task = await self.fetch_task(task_id, version)
            await self.storage.store_task(task, commit_id=version.commit_id)
            return task
        else:
            # Always fetch branches to get latest
            return await self.fetch_task(task_id, version)

    async def generate_task_models(
        self,
        task_id: str | UUID,
        branch_name: str | None = None,
        commit_id: str | None = None,
        output_dir: Path | str | None = None,
    ) -> DatamodelCodegenResponse:
        """
        Generate Python models from all schemas in a task using the new task-based endpoint.

        Args:
            task_id: The task ID to generate models from
            branch_name: Branch name (mutually exclusive with commit_id, defaults to "main")
            commit_id: Commit ID (mutually exclusive with branch_name)
            output_dir: Optional directory to write the generated code to

        Returns:
            DatamodelCodegenResponse object containing the generated code

        Raises:
            RuntimeError: If the generation request fails
            IOError: If file operations fail
        """
        version = VersionRef(branch_name=branch_name, commit_id=commit_id)

        task_id_str = str(task_id) if isinstance(task_id, UUID) else task_id
        logger.info(
            f"Generating task models for task {task_id_str} on {version.model_dump_json()}"
        )

        try:
            # Make POST request to the new task-based endpoint
            params = {}
            if version.branch_name:
                params["branch_name"] = version.branch_name
            else:
                assert version.commit_id is not None
                params["commit_id"] = version.commit_id

            response = await self.post(
                f"/tasks/{task_id_str}/generate-models",
                params=params,
            )

            if response.status_code != 200:
                logger.error(f"Task model generation failed: {response.status_code}")
                logger.error(f"Response content: {response.text}")

                try:
                    error_detail = response.json()
                    error_message = error_detail.get("message", response.text)
                except json.JSONDecodeError:
                    error_message = response.text or f"HTTP {response.status_code}"

                raise RuntimeError(f"Task model generation failed: {error_message}")

            # Parse response
            response_data = self._unwrap_gateway_response(response.json())
            codegen_response = DatamodelCodegenResponse.model_validate(response_data)
            logger.info(f"Generated code for {codegen_response.filename}")

            # Validate generated code
            if not codegen_response.generated_code.strip():
                raise ValueError("Generated code is empty")

            # Optionally write to files
            if output_dir is not None:
                output_path = (
                    Path(output_dir) if isinstance(output_dir, str) else output_dir
                )
                output_path.mkdir(parents=True, exist_ok=True)

                # Sanitize filename to ensure valid Python module name
                sanitized_filename = self._sanitize_module_name(
                    codegen_response.filename
                )
                file_path = output_path / sanitized_filename
                file_path.write_text(codegen_response.generated_code)
                logger.info(f"Saved generated code to {file_path}")

                # Try to format the file with ruff (best-effort, won't fail if ruff unavailable)
                self._format_with_ruff(file_path)

            return codegen_response

        except httpx.TimeoutException as e:
            logger.error(f"Task model generation timed out: {e}")
            raise RuntimeError("Task model generation timed out") from e
        except Exception as e:
            logger.error(f"Task model generation failed: {e}", exc_info=True)
            raise RuntimeError("Task model generation failed") from e

    async def create_task(
        self,
        name: str,
        description: str | None = None,
        branch_name: str = "main",
    ) -> Task:
        """
        Create a new task.

        Args:
            name: Task name (1-255 chars)
            description: Optional description (max 1000 chars)
            branch_name: Branch to create on (defaults to "main")

        Returns:
            The created Task with generated IDs and default branch

        Example:
            task = await client.create_task("Customer Support Bot")
        """
        request = TaskCreateRequest(
            name=name,
            description=description,
            branchName=branch_name,
        )

        response = await self.post(
            "/tasks",
            json=request.model_dump(by_alias=True, exclude_none=True, mode="json"),
        )
        response.raise_for_status()

        task_data = self._unwrap_gateway_response(response.json())
        return Task.model_validate(task_data)

    async def create_prompt(
        self,
        task_id: str | UUID,
        name: str,
        messages: list[MessageData | dict],
        description: str | None = None,
        branch_name: str = "main",
    ) -> PromptTemplate:
        """
        Create a new prompt with messages.

        Args:
            task_id: Parent task ID
            name: Prompt name
            messages: List of MessageData objects or dicts
            description: Optional description
            branch_name: Branch to create on (defaults to "main")

        Returns:
            The created PromptTemplate with auto-generated input schema

        Example:
            from moxn.types.requests import MessageData
            from moxn.types.content import MessageRole
            from moxn.types.blocks.text import TextContentModel
            from moxn.types.blocks.variable import TextVariableModel, VariableFormat

            messages = [
                MessageData(
                    name="System",
                    role=MessageRole.SYSTEM,
                    blocks=[[TextContentModel(text="You are helpful.")]]
                ),
                MessageData(
                    name="User Query",
                    role=MessageRole.USER,
                    blocks=[[TextVariableModel(
                        name="query",
                        format=VariableFormat.INLINE,
                        description="User's question",
                        required=True
                    )]]
                )
            ]
            prompt = await client.create_prompt(task.id, "Q&A", messages)
        """
        task_id = UUID(task_id) if isinstance(task_id, str) else task_id

        # Convert dicts to MessageData if needed
        message_data = [
            MessageData(**msg) if isinstance(msg, dict) else msg for msg in messages
        ]

        request = PromptCreateRequest(
            name=name,
            taskId=task_id,
            description=description,
            branchName=branch_name,
            messages=message_data,
        )

        response = await self.post(
            "/prompts",
            json=request.model_dump(by_alias=True, exclude_none=True, mode="json"),
        )
        response.raise_for_status()

        prompt_data = self._unwrap_gateway_response(response.json())
        prompt = PromptTemplate.model_validate(prompt_data)

        # Register any signed content for auto-refresh
        await self._register_signed_content_in_messages(prompt.messages)

        return prompt

    async def create_telemetry_log(
        self,
        log_request: TelemetryLogRequest,
    ) -> TelemetryLogResponse:
        """Send telemetry log to the API."""
        logger.debug(f"Sending telemetry log: {log_request}")

        try:
            # Ensure proper serialization of all fields
            json_data = log_request.model_dump(
                exclude_none=True,
                mode="json",
                by_alias=True,
            )

            response = await self.client.post(
                "/telemetry/log-event",
                json=json_data,
            )
            response.raise_for_status()

            response_data = self._unwrap_gateway_response(response.json())
            return TelemetryLogResponse.model_validate(response_data)

        except httpx.TimeoutException as e:
            logger.error(f"Telemetry prompt timed out: {e}")
            raise RuntimeError("Telemetry prompt timed out") from e
        except Exception as e:
            logger.error(f"Telemetry prompt failed: {e}", exc_info=True)
            raise RuntimeError("Failed to send telemetry log") from e

    async def create_prompt_session(
        self,
        prompt_id: str,
        branch_name: str | None = None,
        commit_id: str | None = None,
        session_data: RenderableModel | None = None,
    ) -> PromptSession:
        """
        Create a new PromptSession for managing LLM interactions.

        Args:
            prompt_id: The base prompt ID
            branch_name: Branch name (mutually exclusive with commit_id, defaults to "main")
            commit_id: Commit ID (mutually exclusive with branch_name)
            session_data: Optional session data to use in message rendering
        """
        prompt = await self.get_prompt(
            prompt_id, branch_name=branch_name, commit_id=commit_id
        )
        session = PromptSession.from_prompt_template(
            prompt=prompt, session_data=session_data
        )
        return session

    async def prompt_session_from_session_data(
        self,
        session_data: RenderableModel,
        branch_name: str | None = None,
        commit_id: str | None = None,
    ) -> PromptSession:
        """
        Create a new PromptSession for managing LLM interactions.

        Args:
            session_data: Input schema to use in message rendering
            branch_name: Branch name (mutually exclusive with commit_id)
            commit_id: Commit ID (mutually exclusive with branch_name).
                If both branch_name and commit_id are None, falls back to
                session_data.moxn_schema_metadata.commit_id.

        Raises:
            ValueError: If session_data does not have a prompt_id (e.g., task-level schema)
        """
        if not session_data.moxn_schema_metadata.prompt_id:
            raise ValueError(
                "Session data must have prompt_id to create a prompt session. "
                "Task-level schemas without prompt_id cannot be used for sessions."
            )

        # Fallback to session_data's commit_id if no explicit version args
        # Branch-based access is always explicit (no fallback to branch_id)
        effective_commit_id = commit_id
        if branch_name is None and commit_id is None:
            effective_commit_id = session_data.moxn_schema_metadata.commit_id

        prompt = await self.get_prompt(
            session_data.moxn_schema_metadata.prompt_id,
            branch_name=branch_name,
            commit_id=effective_commit_id,
        )
        session = PromptSession.from_prompt_template(
            prompt=prompt, session_data=session_data
        )
        return session

    @asynccontextmanager
    async def span(
        self,
        prompt_session: PromptSession,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
        *,
        parent_context: SpanContext | None = None,
        trace_context: TraceContext | None = None,
    ) -> AsyncGenerator[Span, None]:
        """
        Creates a new span and sets it as the current span.

        Context resolution order:
        1. parent_context parameter (explicit parent for callbacks)
        2. current_span ContextVar (automatic nesting)
        3. trace_context parameter (new root in existing trace)
        4. Generate new trace

        Args:
            prompt_session: The prompt session for this span
            name: Optional span name (defaults to prompt name)
            metadata: User-provided searchable metadata (e.g., customer_id, request_id)
            parent_context: Explicit parent SpanContext (for async/callback patterns)
            trace_context: Explicit TraceContext (for distributed tracing)

        Example:
            async with client.span(
                prompt_session=session,
                name="document_processing",
                metadata={"customer_id": "cust_123", "document_type": "invoice"}
            ) as span:
                # Add more metadata during execution
                span.set_attribute("status", "processing")
                await client.log_telemetry_event(llm_event)

        Example (distributed tracing):
            # Continue a trace from incoming HTTP headers
            trace_ctx = extract_trace_context(request.headers)
            async with client.span(session, trace_context=trace_ctx) as span:
                await client.log_telemetry_event(llm_event)

        Example (async callback):
            # Pass parent context to a callback
            carrier = client.extract_context()
            await queue.put({"carrier": carrier.model_dump(), "data": ...})

            # In callback handler:
            carrier = MoxnTraceCarrier.model_validate(message["carrier"])
            async with client.span_from_carrier(carrier) as span:
                await process(message["data"])
        """
        if name is None:
            name = prompt_session.prompt.name
        async with self.telemetry_client.span(
            prompt_session=prompt_session,
            name=name,
            metadata=metadata,
            parent_context=parent_context,
            trace_context=trace_context,
        ) as span:
            yield span

    @asynccontextmanager
    async def span_from_carrier(
        self,
        carrier: MoxnTraceCarrier,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> AsyncGenerator[Span, None]:
        """
        Create a span that continues from a MoxnTraceCarrier.

        Use this when receiving trace context from another service
        or resuming a trace in a callback/queue handler.

        Args:
            carrier: The trace carrier containing context and Moxn metadata
            name: Optional span name (defaults to prompt name from carrier)
            metadata: User-provided searchable attributes

        Example:
            # In a queue worker or callback handler
            carrier = MoxnTraceCarrier.model_validate(message["carrier"])
            async with client.span_from_carrier(carrier, name="process_item") as span:
                span.set_attribute("item_id", message["item_id"])
                await process(message["data"])
        """
        async with self.telemetry_client.span_from_carrier(
            carrier=carrier,
            name=name,
            metadata=metadata,
        ) as span:
            yield span

    def extract_context(self) -> MoxnTraceCarrier | None:
        """
        Extract current span as a carrier for propagation.

        Use this to pass trace context across async boundaries,
        to queue workers, or to other services.

        Returns:
            MoxnTraceCarrier if there's an active span, None otherwise.

        Example:
            async with client.span(session) as span:
                # Extract carrier to pass to a queue worker
                carrier = client.extract_context()
                if carrier:
                    await queue.put({
                        "carrier": carrier.model_dump(mode="json"),
                        "data": payload
                    })
        """
        return self.telemetry_client.extract_context()

    async def log_telemetry_event(self, event: LLMEvent) -> None:
        """Log an LLM event to the current span context or create a temporary span if needed."""
        event = event.model_copy(deep=True)
        await self.telemetry_client.log_event(event=event)

    @overload
    async def log_telemetry_event_from_response(
        self,
        prompt_session: PromptSession,
        response: AnthropicMessage,
        provider: Literal[Provider.ANTHROPIC],
    ) -> None: ...

    @overload
    async def log_telemetry_event_from_response(
        self,
        prompt_session: PromptSession,
        response: OpenAIChatCompletion,
        provider: Literal[Provider.OPENAI_CHAT],
    ) -> None: ...

    @overload
    async def log_telemetry_event_from_response(
        self,
        prompt_session: PromptSession,
        response: GoogleGenerateContentResponse,
        provider: Literal[Provider.GOOGLE_GEMINI] | Literal[Provider.GOOGLE_VERTEX],
    ) -> None: ...

    async def log_telemetry_event_from_response(
        self,
        prompt_session: PromptSession,
        response: (
            OpenAIChatCompletion | AnthropicMessage | GoogleGenerateContentResponse
        ),
        provider: Provider,
    ) -> None:
        """Log an LLM event to the current span context."""
        match provider:
            case Provider.ANTHROPIC:
                event = prompt_session.create_llm_event_from_response(
                    response=cast(AnthropicMessage, response),
                    provider=provider,
                )
            case Provider.OPENAI_CHAT:
                event = prompt_session.create_llm_event_from_response(
                    response=cast(OpenAIChatCompletion, response),
                    provider=provider,
                )
            case Provider.GOOGLE_GEMINI:
                event = prompt_session.create_llm_event_from_response(
                    response=cast(GoogleGenerateContentResponse, response),
                    provider=provider,
                )
            case Provider.GOOGLE_VERTEX:
                event = prompt_session.create_llm_event_from_response(
                    response=cast(GoogleGenerateContentResponse, response),
                    provider=provider,
                )
            case _:
                raise ValueError(f"Unsupported provider: {provider}")

        await self.log_telemetry_event(event=event)

    async def flush(self, timeout: float | None = None) -> None:
        """
        Await in-flight telemetry logs (call this at process-exit,
        lambda return, or FASTAPI shutdown).
        """
        if timeout is None:
            timeout = self.settings.telemetry_timeout
        await self.telemetry_client._dispatcher.flush(timeout)


def _get_headers(settings: MoxnSettings, _: bool = True) -> dict:
    """Returns headers for API requests - tenant-scoped auth (API key only)."""
    return {
        "x-api-key": settings.api_key.get_secret_value(),
        "Content-Type": "application/json",
    }



"""Structured output formatting utilities for converting schemas to provider-specific formats.

This module provides formatters for converting SdkSchema to provider-specific
structured output / response format configurations.

Provider Translation Reference:

| Provider   | Response Format                                                              | Notes                                              |
|------------|------------------------------------------------------------------------------|----------------------------------------------------|
| OpenAI     | response_format: {type: 'json_schema', json_schema: {name, schema, strict}}  | Requires parallel_tool_calls: false                |
| Anthropic  | output_format: {type: 'json_schema', schema}                                 | Requires anthropic-beta: structured-outputs header |
| Google     | generation_config: {response_mime_type: 'application/json', response_schema} |                                                    |
"""

from __future__ import annotations

import json
from typing import Any

from moxn.types.tool import SdkSchema


class StructuredOutputFormatter:
    """Converts SdkSchema to provider-specific structured output formats."""

    @staticmethod
    def to_openai_response_format(schema: SdkSchema) -> dict[str, Any]:
        """Build OpenAI response_format with json_schema.

        OpenAI format:
        {
            "type": "json_schema",
            "json_schema": {
                "name": "schema_name",
                "schema": { ... JSON Schema ... },
                "strict": true
            }
        }

        Note: When using this, parallel_tool_calls must be set to false.

        Args:
            schema: SdkSchema containing the JSON Schema

        Returns:
            OpenAI response_format configuration
        """
        return {
            "type": "json_schema",
            "json_schema": {
                "name": schema.name,
                "schema": json.loads(schema.exported_json),
                "strict": True,
            },
        }

    @staticmethod
    def to_anthropic_output_format(schema: SdkSchema) -> dict[str, Any]:
        """Build Anthropic output_format for structured outputs.

        Anthropic format:
        {
            "type": "json_schema",
            "schema": { ... JSON Schema ... }
        }

        Note: This requires the beta header:
            extra_headers={"anthropic-beta": "structured-outputs-2025-11-13"}

        Args:
            schema: SdkSchema containing the JSON Schema

        Returns:
            Anthropic output_format configuration
        """
        return {
            "type": "json_schema",
            "schema": json.loads(schema.exported_json),
        }

    @staticmethod
    def to_google_generation_config(schema: SdkSchema) -> dict[str, Any]:
        """Build Google generation_config for structured output.

        Google format (partial generation_config):
        {
            "response_mime_type": "application/json",
            "response_schema": { ... JSON Schema ... }
        }

        Note: This should be merged with other generation_config parameters.

        Args:
            schema: SdkSchema containing the JSON Schema

        Returns:
            Google generation_config parameters for structured output
        """
        return {
            "response_mime_type": "application/json",
            "response_schema": json.loads(schema.exported_json),
        }



from typing import Any, Literal, overload

from moxn.base_models.blocks.base import ToProviderContentBlockMixin
from moxn.base_models.blocks.context import MessageContext
from moxn.types.blocks.text import (
    ReasoningContentModel,
    TextContentModel,
    ThinkingContentModel,
)
from moxn.types.content import Provider
from moxn.types.type_aliases.anthropic import AnthropicTextBlockParam
from moxn.types.type_aliases.google import GooglePart
from moxn.types.type_aliases.openai_chat import (
    OpenAIChatCompletionContentPartTextParam,
)
from moxn.types.type_aliases.openai_responses import (
    OpenAIResponsesInputTextParam,
)


class ToProviderContentBlockMixinTextOnly(
    ToProviderContentBlockMixin
):  # Override the to_provider_content_block method with more specific types
    @overload
    def to_provider_content_block(
        self, provider: Literal[Provider.ANTHROPIC], context: MessageContext
    ) -> AnthropicTextBlockParam: ...

    @overload
    def to_provider_content_block(
        self, provider: Literal[Provider.OPENAI_CHAT], context: MessageContext
    ) -> OpenAIChatCompletionContentPartTextParam: ...

    @overload
    def to_provider_content_block(
        self, provider: Literal[Provider.OPENAI_RESPONSES], context: MessageContext
    ) -> OpenAIResponsesInputTextParam: ...

    @overload
    def to_provider_content_block(
        self, provider: Literal[Provider.GOOGLE_GEMINI], context: MessageContext
    ) -> GooglePart: ...

    @overload
    def to_provider_content_block(
        self, provider: Literal[Provider.GOOGLE_VERTEX], context: MessageContext
    ) -> GooglePart: ...

    # Implement the method (same as in parent class)
    def to_provider_content_block(self, provider: Provider, context: MessageContext):
        if provider == Provider.ANTHROPIC:
            return self._to_anthropic_content_block(context)
        elif provider == Provider.OPENAI_CHAT:
            return self._to_openai_chat_content_block(context)
        elif provider == Provider.OPENAI_RESPONSES:
            return self._to_openai_responses_content_block(context)
        elif provider == Provider.GOOGLE_GEMINI:
            return self._to_google_gemini_content_block(context)
        elif provider == Provider.GOOGLE_VERTEX:
            return self._to_google_vertex_content_block(context)
        else:
            raise ValueError("Unsupported provider")

    def _to_anthropic_content_block(
        self, context: MessageContext
    ) -> AnthropicTextBlockParam:
        raise NotImplementedError

    def _to_openai_chat_content_block(
        self, context: MessageContext
    ) -> OpenAIChatCompletionContentPartTextParam:
        raise NotImplementedError

    def _to_openai_responses_content_block(
        self, context: MessageContext
    ) -> OpenAIResponsesInputTextParam:
        raise NotImplementedError

    def _to_google_gemini_content_block(self, context: MessageContext) -> GooglePart:
        raise NotImplementedError

    def _to_google_vertex_content_block(self, context: MessageContext) -> GooglePart:
        raise NotImplementedError


class TextContent(TextContentModel, ToProviderContentBlockMixinTextOnly):
    """Text content block."""

    text: str

    def _to_anthropic_content_block(
        self, context: MessageContext
    ) -> AnthropicTextBlockParam:
        return AnthropicTextBlockParam(type="text", text=self.text)

    def _to_openai_chat_content_block(
        self, context: MessageContext
    ) -> OpenAIChatCompletionContentPartTextParam:
        return OpenAIChatCompletionContentPartTextParam(type="text", text=self.text)

    def _to_openai_responses_content_block(
        self, context: MessageContext
    ) -> OpenAIResponsesInputTextParam:
        return OpenAIResponsesInputTextParam(type="input_text", text=self.text)

    def _to_google_gemini_content_block(self, context: MessageContext) -> GooglePart:
        return GooglePart.from_text(text=self.text)

    def _to_google_vertex_content_block(self, context: MessageContext) -> GooglePart:
        return GooglePart.from_text(text=self.text)


class ThinkingContent(ThinkingContentModel, ToProviderContentBlockMixinTextOnly):
    """Thinking content block with bidirectional provider support.

    Used for Claude's extended thinking and Gemini's thinking parts.
    When round-tripping to providers that don't support thinking blocks,
    the content is converted to a text block with a [Thinking] prefix.
    """

    thinking: str

    def _to_anthropic_content_block(
        self, context: MessageContext
    ) -> dict[str, Any]:
        # Anthropic supports thinking blocks natively
        return {"type": "thinking", "thinking": self.thinking}

    def _to_openai_chat_content_block(
        self, context: MessageContext
    ) -> OpenAIChatCompletionContentPartTextParam:
        # OpenAI doesn't support thinking in requests - convert to text
        return OpenAIChatCompletionContentPartTextParam(
            type="text", text=f"[Thinking]\n{self.thinking}"
        )

    def _to_openai_responses_content_block(
        self, context: MessageContext
    ) -> OpenAIResponsesInputTextParam:
        # OpenAI Responses doesn't support thinking in requests - convert to text
        return OpenAIResponsesInputTextParam(
            type="input_text", text=f"[Thinking]\n{self.thinking}"
        )

    def _to_google_gemini_content_block(self, context: MessageContext) -> GooglePart:
        # Google may support thinking parts - convert to text for now
        return GooglePart.from_text(text=f"[Thinking]\n{self.thinking}")

    def _to_google_vertex_content_block(self, context: MessageContext) -> GooglePart:
        return GooglePart.from_text(text=f"[Thinking]\n{self.thinking}")


class ReasoningContent(ReasoningContentModel, ToProviderContentBlockMixinTextOnly):
    """Reasoning content block for OpenAI o1/o3 model reasoning summaries.

    When round-tripping to providers, the reasoning summary is converted
    to a text block with a [Reasoning] prefix.
    """

    summary: str

    def _to_anthropic_content_block(
        self, context: MessageContext
    ) -> AnthropicTextBlockParam:
        # Anthropic doesn't have reasoning - convert to text
        return AnthropicTextBlockParam(
            type="text", text=f"[Reasoning]\n{self.summary}"
        )

    def _to_openai_chat_content_block(
        self, context: MessageContext
    ) -> OpenAIChatCompletionContentPartTextParam:
        # Reasoning is output-only in o1/o3 - convert to text for round-trip
        return OpenAIChatCompletionContentPartTextParam(
            type="text", text=f"[Reasoning]\n{self.summary}"
        )

    def _to_openai_responses_content_block(
        self, context: MessageContext
    ) -> OpenAIResponsesInputTextParam:
        return OpenAIResponsesInputTextParam(
            type="input_text", text=f"[Reasoning]\n{self.summary}"
        )

    def _to_google_gemini_content_block(self, context: MessageContext) -> GooglePart:
        return GooglePart.from_text(text=f"[Reasoning]\n{self.summary}")

    def _to_google_vertex_content_block(self, context: MessageContext) -> GooglePart:
        return GooglePart.from_text(text=f"[Reasoning]\n{self.summary}")



from typing import Any, Generic, Literal, TypeVar

from pydantic import BaseModel, ConfigDict, Field

from moxn.types.blocks.base import BlockType
from moxn.types.blocks.image import ImageContentFromSourceModel
from moxn.types.blocks.text import TextContentModel

T = TypeVar("T")


class ToolCallModel(BaseModel):
    block_type: Literal[BlockType.TOOL_CALL] = Field(
        default=BlockType.TOOL_CALL, alias="blockType"
    )
    id: str
    arguments: str | dict[str, Any] | None
    name: str

    model_config = ConfigDict(populate_by_name=True)


class ToolResultBase(BaseModel, Generic[T]):
    block_type: Literal[BlockType.TOOL_RESULT] = Field(
        default=BlockType.TOOL_RESULT, alias="blockType"
    )
    type: Literal["tool_use"]
    id: str
    name: str
    content: T | None

    model_config = ConfigDict(populate_by_name=True)


class ToolResultModel(
    ToolResultBase[TextContentModel | ImageContentFromSourceModel | None]
):
    type: Literal["tool_use"]
    id: str
    name: str
    content: TextContentModel | ImageContentFromSourceModel | None



"""Content block model for DTOs - uses flexible RawContentBlock for API parsing."""

from pydantic import TypeAdapter
from moxn.types.blocks.raw_block import RawContentBlock

# For DTOs, we use RawContentBlock to ensure flexible API parsing
ContentBlockModel = RawContentBlock

ContentBlockAdapter: TypeAdapter[ContentBlockModel] = TypeAdapter(ContentBlockModel)



"""Block conversion between raw API blocks and typed domain models."""

from pathlib import Path
from typing import Any

from moxn.types.blocks.raw_block import RawContentBlock
from moxn.types.blocks.typed_blocks import (
    BlockType,
    FileContent,
    FileSourceType,
    ImageContent,
    ImageSourceType,
    SignedURLContent,
    TextContent,
    ToolCallContent,
    ToolResultContent,
    TypedContentBlock,
    VariableContent,
    VariableFormat,
    VariableType,
)


class BlockConversionError(Exception):
    """Raised when block conversion fails."""

    pass


def to_typed_block(raw: RawContentBlock) -> TypedContentBlock:
    """
    Convert a raw content block from API to a typed domain model.

    Args:
        raw: Raw content block from API

    Returns:
        Typed content block for use in provider conversions

    Raises:
        BlockConversionError: If conversion fails
    """
    try:
        match raw.block_type:
            case BlockType.TEXT:
                return _convert_text_block(raw)
            case BlockType.VARIABLE:
                return _convert_variable_block(raw)
            case BlockType.IMAGE:
                return _convert_image_block(raw)
            case BlockType.FILE:
                return _convert_file_block(raw)
            case BlockType.SIGNED:
                return _convert_signed_block(raw)
            case BlockType.TOOL_CALL:
                return _convert_tool_call_block(raw)
            case BlockType.TOOL_RESULT:
                return _convert_tool_result_block(raw)
            case _:
                raise BlockConversionError(f"Unknown block type: {raw.block_type}")

    except Exception as e:
        raise BlockConversionError(
            f"Failed to convert {raw.block_type} block: {e}"
        ) from e


def from_typed_block(typed: TypedContentBlock) -> RawContentBlock:
    """
    Convert a typed domain model back to a raw content block for API.

    Args:
        typed: Typed content block

    Returns:
        Raw content block for API serialization
    """
    # Create base raw block
    raw_data: dict[str, Any] = {
        "blockType": typed.block_type.value,
        "options": typed.options,
    }

    if isinstance(typed, TextContent):
        raw_data["text"] = typed.text

    elif isinstance(typed, VariableContent):
        raw_data.update(
            {
                "name": typed.name,
                "variableType": typed.variable_type.value,
                "format": typed.format.value,
                "description": typed.description,
                "required": typed.required,
            }
        )
        if typed.default_value is not None:
            raw_data["defaultValue"] = typed.default_value

    elif isinstance(typed, ImageContent):
        raw_data.update(
            {
                "type": typed.source_type.value,
                "mediaType": typed.media_type,
            }
        )
        # Add source-specific fields
        if typed.base64:
            raw_data["base64"] = typed.base64
        elif typed.url:
            raw_data["url"] = typed.url
        elif typed.filepath:
            raw_data["filepath"] = str(typed.filepath)
        elif typed.google_uri:
            raw_data["uri"] = typed.google_uri
        # Note: bytes and google_file are not serializable to API

    elif isinstance(typed, FileContent):
        raw_data.update(
            {
                "type": typed.source_type.value,
                "mediaType": typed.media_type,
            }
        )
        # Add source-specific fields
        if typed.base64:
            raw_data["base64"] = typed.base64
        elif typed.url:
            raw_data["url"] = typed.url
        elif typed.openai_file_id:
            raw_data["openaiFileId"] = typed.openai_file_id
        elif typed.google_uri:
            raw_data["uri"] = typed.google_uri

    elif isinstance(typed, SignedURLContent):
        raw_data["filePath"] = typed.file_path
        if typed.media_type:
            raw_data["mediaType"] = typed.media_type

    elif isinstance(typed, ToolCallContent):
        raw_data.update(
            {
                "toolName": typed.tool_name,
                "toolCallId": typed.tool_call_id,
                "arguments": typed.arguments,
            }
        )

    elif isinstance(typed, ToolResultContent):
        raw_data.update(
            {
                "toolCallId": typed.tool_call_id,
                "result": typed.result,
            }
        )

    else:
        raise BlockConversionError(f"Unknown block type: {typed.block_type}")

    return RawContentBlock(**raw_data)


def _convert_text_block(raw: RawContentBlock) -> TextContent:
    """Convert raw text block to typed text content."""
    if not raw.text:
        raise BlockConversionError("Text block missing required 'text' field")

    return TextContent(
        text=raw.text,
        options=raw.options,
    )


def _convert_variable_block(raw: RawContentBlock) -> VariableContent:
    """Convert raw variable block to typed variable content."""
    if not raw.name:
        raise BlockConversionError("Variable block missing required 'name' field")
    if not raw.variable_type:
        raise BlockConversionError(
            "Variable block missing required 'variable_type' field"
        )
    if not raw.format:
        raise BlockConversionError("Variable block missing required 'format' field")

    try:
        variable_type = VariableType(raw.variable_type)
        format_type = VariableFormat(raw.format)
    except ValueError as e:
        raise BlockConversionError(f"Invalid variable type or format: {e}") from e

    return VariableContent(
        name=raw.name,
        variable_type=variable_type,
        format=format_type,
        description=raw.description or "",
        required=raw.required if raw.required is not None else True,
        default_value=raw.default_value,
        options=raw.options,
    )


def _convert_image_block(raw: RawContentBlock) -> ImageContent:
    """Convert raw image block to typed image content."""
    if not raw.type:
        raise BlockConversionError("Image block missing required 'type' field")
    if not raw.media_type:
        raise BlockConversionError("Image block missing required 'media_type' field")

    try:
        source_type = ImageSourceType(raw.type)
    except ValueError as e:
        raise BlockConversionError(f"Invalid image source type: {e}") from e

    # Validate required fields based on source type
    image_content = ImageContent(
        source_type=source_type,
        media_type=raw.media_type,
        options=raw.options,
    )

    # Set source-specific fields
    match source_type:
        case ImageSourceType.BASE64:
            if not raw.base64:
                raise BlockConversionError("base64 image missing 'base64' field")
            image_content.base64 = raw.base64
        case ImageSourceType.BYTES:
            if not raw.bytes_data:
                raise BlockConversionError("bytes image missing 'bytes_data' field")
            image_content.bytes_data = raw.bytes_data
        case ImageSourceType.URL:
            if not raw.url:
                raise BlockConversionError("URL image missing 'url' field")
            image_content.url = raw.url
        case ImageSourceType.LOCAL_FILE:
            if not raw.filepath:
                raise BlockConversionError("local file image missing 'filepath' field")
            image_content.filepath = Path(raw.filepath)
        case ImageSourceType.GOOGLE_FILE_REFERENCE:
            if not raw.uri:
                raise BlockConversionError("Google file reference missing 'uri' field")
            image_content.google_uri = raw.uri
        # GOOGLE_FILE requires the actual file object, handled elsewhere

    return image_content


def _convert_file_block(raw: RawContentBlock) -> FileContent:
    """Convert raw file block to typed file content."""
    if not raw.type:
        raise BlockConversionError("File block missing required 'type' field")
    if not raw.media_type:
        raise BlockConversionError("File block missing required 'media_type' field")

    try:
        source_type = FileSourceType(raw.type)
    except ValueError as e:
        raise BlockConversionError(f"Invalid file source type: {e}") from e

    file_content = FileContent(
        source_type=source_type,
        media_type=raw.media_type,
        options=raw.options,
    )

    # Set source-specific fields
    match source_type:
        case FileSourceType.BASE64:
            if not raw.base64:
                raise BlockConversionError("base64 file missing 'base64' field")
            file_content.base64 = raw.base64
        case FileSourceType.BYTES:
            if not raw.bytes_data:
                raise BlockConversionError("bytes file missing 'bytes_data' field")
            file_content.bytes_data = raw.bytes_data
        case FileSourceType.URL:
            if not raw.url:
                raise BlockConversionError("URL file missing 'url' field")
            file_content.url = raw.url
        case FileSourceType.GOOGLE_FILE_REFERENCE:
            if not raw.uri:
                raise BlockConversionError("Google file reference missing 'uri' field")
            file_content.google_uri = raw.uri
        # Other types handled elsewhere

    return file_content


def _convert_signed_block(raw: RawContentBlock) -> SignedURLContent:
    """Convert raw signed URL block to typed signed URL content."""
    if not raw.file_path:
        raise BlockConversionError(
            "Signed URL block missing required 'file_path' field"
        )

    return SignedURLContent(
        file_path=raw.file_path,
        media_type=raw.media_type,
        options=raw.options,
    )


def _convert_tool_call_block(raw: RawContentBlock) -> ToolCallContent:
    """Convert raw tool call block to typed tool call content."""
    if not raw.tool_name:
        raise BlockConversionError("Tool call block missing required 'tool_name' field")
    if not raw.tool_call_id:
        raise BlockConversionError(
            "Tool call block missing required 'tool_call_id' field"
        )

    return ToolCallContent(
        tool_name=raw.tool_name,
        tool_call_id=raw.tool_call_id,
        arguments=getattr(raw, "arguments", {}),
        options=raw.options,
    )


def _convert_tool_result_block(raw: RawContentBlock) -> ToolResultContent:
    """Convert raw tool result block to typed tool result content."""
    if not raw.tool_call_id:
        raise BlockConversionError(
            "Tool result block missing required 'tool_call_id' field"
        )

    return ToolResultContent(
        tool_call_id=raw.tool_call_id,
        result=getattr(raw, "result", None),
        options=raw.options,
    )



"""Request configuration models for capturing provider-specific LLM call parameters.

These models capture the configuration used when making requests to LLM providers,
enabling better observability and debugging of structured generation and tool calling.
"""

from typing import Any, Literal, Type

from pydantic import BaseModel, ConfigDict, Field, field_serializer

from moxn.types.content import Provider


def _is_pydantic_model(obj: Any) -> bool:
    """Check if an object is a Pydantic model class (not instance)."""
    try:
        return isinstance(obj, type) and issubclass(obj, BaseModel)
    except TypeError:
        return False


def _convert_pydantic_to_schema(value: Any) -> Any:
    """Recursively convert Pydantic model classes to JSON schemas.

    Args:
        value: Any value that might contain Pydantic model classes

    Returns:
        The same value with all Pydantic model classes converted to JSON schemas
    """
    if value is None:
        return None

    # Handle Pydantic model class
    if _is_pydantic_model(value):
        return value.model_json_schema()

    # Handle dict (might contain nested Pydantic models)
    if isinstance(value, dict):
        return {k: _convert_pydantic_to_schema(v) for k, v in value.items()}

    # Handle list (might contain nested Pydantic models)
    if isinstance(value, list):
        return [_convert_pydantic_to_schema(item) for item in value]

    # Return as-is for other types
    return value


class CompletionConfig(BaseModel):
    """Completion configuration stored on a prompt.

    Represents default LLM parameters configured in the Moxn UI.
    These are stored defaults, distinct from runtime RequestConfig which
    captures provider-specific runtime parameters for telemetry.
    """

    provider: Provider | None = None
    model: str | None = None
    temperature: float | None = None
    max_tokens: int | None = Field(None, alias="maxTokens")
    top_p: float | None = Field(None, alias="topP")
    thinking: dict[str, Any] | None = Field(
        None,
        description="Provider-specific thinking/reasoning configuration. "
        "Passed through to provider APIs without modification.",
    )

    # Tool calling configuration
    tool_choice: str | dict[str, Any] | None = Field(
        None,
        alias="toolChoice",
        description="Tool choice behavior: 'auto', 'none', 'required', or {tool: 'name'} "
        "to force a specific tool. Translated to provider-specific format by SDK.",
    )

    # Structured output configuration
    response_schema_id: str | None = Field(
        None,
        alias="responseSchemaId",
        description="UUID of the structured_output schema in prompt.tools to use for "
        "response formatting. SDK looks up the full schema locally (no fetch).",
    )

    # Parallel tool calls (OpenAI constraint)
    parallel_tool_calls: bool | None = Field(
        None,
        alias="parallelToolCalls",
        description="Whether to allow parallel tool calls. Must be False when using "
        "structured outputs with OpenAI (SDK enforces this constraint).",
    )

    model_config = ConfigDict(populate_by_name=True)


class RequestConfig(BaseModel):
    """Base configuration for all providers.

    Captures common parameters shared across LLM providers.
    """

    provider: Provider
    temperature: float | None = None
    max_tokens: int | None = None


class OpenAIChatRequestConfig(RequestConfig):
    """OpenAI Chat API specific configuration.

    Captures parameters specific to OpenAI's Chat Completions API,
    including structured outputs and function calling.
    """

    provider: Literal[Provider.OPENAI_CHAT] = Provider.OPENAI_CHAT
    response_format: dict[str, Any] | None = Field(
        default=None,
        description="Response format config: {type: 'json_object'} or {type: 'json_schema', json_schema: {...}}",
    )
    tools: list[dict[str, Any]] | None = Field(
        default=None, description="Tool/function definitions for function calling"
    )
    tool_choice: str | dict[str, Any] | None = Field(
        default=None,
        description="Tool choice: 'auto', 'none', 'required', or specific tool",
    )
    parallel_tool_calls: bool | None = Field(
        default=None, description="Whether to allow parallel tool calls"
    )
    reasoning_effort: str | None = Field(
        default=None,
        description="Reasoning effort level for o-series and GPT-5 models (none/minimal/low/medium/high)",
    )

    @field_serializer("response_format")
    def serialize_response_format(
        self, value: dict[str, Any] | None
    ) -> dict[str, Any] | None:
        """Convert Pydantic models in response_format to JSON schemas."""
        return _convert_pydantic_to_schema(value)

    @field_serializer("tools")
    def serialize_tools(
        self, value: list[dict[str, Any]] | None
    ) -> list[dict[str, Any]] | None:
        """Convert Pydantic models in tool definitions to JSON schemas."""
        return _convert_pydantic_to_schema(value)


class OpenAIResponsesRequestConfig(RequestConfig):
    """OpenAI Responses API specific configuration.

    Captures parameters specific to OpenAI's Responses API,
    which uses items-based input/output model.
    """

    provider: Literal[Provider.OPENAI_RESPONSES] = Provider.OPENAI_RESPONSES
    tools: list[dict[str, Any]] | None = Field(
        default=None, description="Tool/function definitions for function calling"
    )
    tool_choice: str | dict[str, Any] | None = Field(
        default=None,
        description="Tool choice: 'auto', 'none', 'required', or specific tool",
    )
    parallel_tool_calls: bool | None = Field(
        default=None, description="Whether to allow parallel tool calls"
    )
    reasoning: dict[str, Any] | None = Field(
        default=None,
        description="Reasoning configuration for o-series models (effort, summary)",
    )

    @field_serializer("tools")
    def serialize_tools(
        self, value: list[dict[str, Any]] | None
    ) -> list[dict[str, Any]] | None:
        """Convert Pydantic models in tool definitions to JSON schemas."""
        return _convert_pydantic_to_schema(value)


class AnthropicRequestConfig(RequestConfig):
    """Anthropic Claude specific configuration.

    Captures parameters specific to Anthropic's Messages API,
    including tool use for both function calling and structured generation.
    """

    provider: Literal[Provider.ANTHROPIC] = Provider.ANTHROPIC
    tools: list[dict[str, Any]] | None = Field(
        default=None,
        description="Tool definitions for both tool use and structured generation",
    )
    tool_choice: dict[str, Any] | None = Field(
        default=None, description="Tool choice: {type: 'auto'|'any'|'tool', name: ...}"
    )
    thinking: dict[str, Any] | None = Field(
        default=None,
        description="Extended thinking configuration (type: 'enabled', budget_tokens: int)",
    )

    @field_serializer("tools")
    def serialize_tools(
        self, value: list[dict[str, Any]] | None
    ) -> list[dict[str, Any]] | None:
        """Convert Pydantic models in tool definitions to JSON schemas."""
        return _convert_pydantic_to_schema(value)


class GoogleRequestConfig(RequestConfig):
    """Google Gemini/Vertex specific configuration.

    Captures parameters specific to Google's generative AI APIs,
    including native structured generation via response_schema.
    """

    provider: Literal[Provider.GOOGLE_GEMINI, Provider.GOOGLE_VERTEX]
    response_schema: dict[str, Any] | Type[BaseModel] | None = Field(
        default=None,
        description="OpenAPI 3.0 schema for structured generation (dict or Pydantic model)",
    )
    response_mime_type: str | None = Field(
        default=None, description="MIME type for response (e.g., 'application/json')"
    )
    function_declarations: list[dict[str, Any]] | None = Field(
        default=None, description="Function declarations for function calling"
    )
    tool_config: dict[str, Any] | None = Field(
        default=None, description="Tool configuration (mode, allowed_functions, etc.)"
    )
    thinking_config: dict[str, Any] | None = Field(
        default=None,
        description="Thinking configuration (thinking_budget for Gemini 2.5, thinking_level for Gemini 3)",
    )

    @field_serializer("response_schema")
    def serialize_response_schema(
        self, value: dict[str, Any] | Type[BaseModel] | None
    ) -> dict[str, Any] | None:
        """Convert Pydantic models in response_schema to JSON schemas."""
        return _convert_pydantic_to_schema(value)

    @field_serializer("function_declarations")
    def serialize_function_declarations(
        self, value: list[dict[str, Any]] | None
    ) -> list[dict[str, Any]] | None:
        """Convert Pydantic models in function declarations to JSON schemas."""
        return _convert_pydantic_to_schema(value)


class SchemaDefinition(BaseModel):
    """Captures schema/model used for structured generation or tool definitions.

    This model stores the actual schema or tool definitions used in a request,
    enabling traceability and debugging of structured generation and function calling.
    """

    definition_type: Literal[
        "json_schema", "pydantic", "tools", "functions", "response_schema"
    ]
    definition: dict[str, Any] | list[dict[str, Any]] | Type[BaseModel] = Field(
        description="The actual JSON Schema, tool definitions, or response schema (dict/list for JSON schema, or Pydantic model class)"
    )
    pydantic_model_name: str | None = Field(
        default=None, description="Name of the Pydantic model if one was used"
    )
    is_structured_output: bool = Field(
        default=False,
        description="True if used for structured generation, False if used for tools/functions",
    )

    @field_serializer("definition")
    def serialize_definition(
        self, value: dict[str, Any] | list[dict[str, Any]] | Type[BaseModel]
    ) -> dict[str, Any] | list[dict[str, Any]]:
        """Convert Pydantic models in definition to JSON schemas."""
        return _convert_pydantic_to_schema(value)



from typing import Sequence, Union

# Provider payload types (for prompt conversion)
from moxn.types.type_aliases.anthropic import (
    AnthropicContentBlockParam,
    AnthropicMessage,
    AnthropicMessageParam,
    AnthropicMessagesParam,
    AnthropicTextBlockParam,
)
from moxn.types.type_aliases.google import (
    GoogleContent,
    GoogleContentBlock,
    GoogleGenerateContentResponse,
    GoogleMessagesParam,
)
from moxn.types.type_aliases.openai_chat import (
    OpenAIChatAssistantMessageParam,
    OpenAIChatCompletion,
    OpenAIChatContentBlock,
    OpenAIChatMessagesParam,
    OpenAIChatSystemMessageParam,
    OpenAIChatUserMessageParam,
)
from moxn.types.type_aliases.openai_responses import (
    OpenAIResponse,
    OpenAIResponsesContentBlock,
    OpenAIResponsesInputItemParam,
    OpenAIResponsesMessagesParam,
)
from moxn.types.type_aliases.invocation import (
    AnthropicInvocationParam,
    GoogleInvocationParam,
    OpenAIChatInvocationParam,
    OpenAIResponsesInvocationParam,
    ProviderInvocationPayload,
)

# Content block types
ProviderContentBlock = (
    AnthropicContentBlockParam
    | OpenAIChatContentBlock
    | OpenAIResponsesContentBlock
    | GoogleContentBlock
)

ProviderContentBlockSequence = (
    Sequence[Sequence[AnthropicContentBlockParam]]
    | Sequence[Sequence[OpenAIChatContentBlock]]
    | Sequence[Sequence[OpenAIResponsesContentBlock]]
    | Sequence[Sequence[GoogleContentBlock]]
)

# Provider response types (for parsing)
ProviderResponse = Union[
    AnthropicMessage,
    OpenAIChatCompletion,
    OpenAIResponse,
    GoogleGenerateContentResponse,
]

# Provider message param types (for message conversion)
ProviderMessageParam = Union[
    AnthropicTextBlockParam,
    AnthropicMessageParam,
    OpenAIChatSystemMessageParam,
    OpenAIChatUserMessageParam,
    OpenAIChatAssistantMessageParam,
    OpenAIResponsesInputItemParam,  # Properly typed Responses API items
    GoogleContent,
]


ProviderPayload = Union[
    AnthropicMessagesParam,
    OpenAIChatMessagesParam,
    OpenAIResponsesMessagesParam,
    GoogleMessagesParam,
]



from __future__ import annotations

import json
from datetime import datetime, timezone
from typing import Any, Protocol
from uuid import uuid4

import httpx
from pydantic import BaseModel, HttpUrl, PrivateAttr, SecretStr, TypeAdapter

# TypeAdapter for serializing dicts that may contain Pydantic models
_content_adapter: TypeAdapter[dict[str, Any]] = TypeAdapter(dict[str, Any])

from moxn.settings import MoxnSettings, get_moxn_settings
from moxn.types.telemetry import (
    MAX_INLINE_CONTENT_SIZE,
    SignedURLRequest,
    SignedURLResponse,
    TelemetryLogRequest,
    TelemetryLogResponse,
)


class TelemetryTransportBackend(Protocol):
    """Protocol for the backend that handles actual sending of telemetry data"""

    async def send_telemetry_log(
        self, log_request: TelemetryLogRequest
    ) -> TelemetryLogResponse:
        """Send a telemetry log request"""
        ...

    async def aclose(self) -> None:
        """Close the backend and cleanup resources"""
        ...


class HttpTelemetryBackend(BaseModel):
    """
    Concrete implementation that talks to the Moxn API over HTTP,
    reusing a single AsyncClient for all telemetry posts and
    a second anonymous client for external-attributes PUTs.
    """

    api_key: SecretStr
    base_url: HttpUrl

    timeout: float = get_moxn_settings().telemetry_timeout

    # these clients are not part of the pydantic model, just PrivateAttrs
    _client: httpx.AsyncClient = PrivateAttr()
    _anon_client: httpx.AsyncClient = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        # 1) main client with base_url and auth headers baked in
        self._client = httpx.AsyncClient(
            base_url=str(self.base_url),
            timeout=self.timeout,
            headers=self.get_headers(),
        )
        # 2) anonymous client for external uploads (no base_url, no auth headers)
        self._anon_client = httpx.AsyncClient(timeout=self.timeout)

    @classmethod
    def from_settings(cls, settings: MoxnSettings) -> "HttpTelemetryBackend":
        return cls(
            base_url=settings.base_api_route,
            api_key=settings.api_key,
            timeout=settings.telemetry_timeout,
        )

    def get_headers(self) -> dict[str, str]:
        return {
            "x-api-key": self.api_key.get_secret_value(),
            "Content-Type": "application/json",
        }

    # ---------- public-protocol methods --------------------------------------

    async def send_telemetry_log(
        self, log_request: TelemetryLogRequest
    ) -> TelemetryLogResponse:
        """Send a telemetry log, delegating content to storage if too large"""

        # Check if content should be stored externally
        if log_request.content and self._should_use_external_storage(
            log_request.content
        ):
            return await self._send_telemetry_log_with_external_content(log_request)

        # inline case: use the persistent client with base_url
        resp = await self._client.post(
            "/telemetry/log-event",
            json=log_request.model_dump(exclude_none=True, mode="json", by_alias=True),
        )
        resp.raise_for_status()
        return TelemetryLogResponse.model_validate(resp.json())

    async def _send_telemetry_log_and_get_signed_url(
        self, req: SignedURLRequest
    ) -> SignedURLResponse:
        resp = await self._client.post(
            "/telemetry/log-event-and-get-signed-url",
            json=req.model_dump(exclude_none=True, mode="json", by_alias=True),
        )
        resp.raise_for_status()
        return SignedURLResponse.model_validate(resp.json())

    # ---------- internal helpers ---------------------------------------------

    def _extract_tenant_id_from_api_key(self) -> str:
        """Extract tenant ID from API key format: moxn_{tenant_id}.{rest}"""
        api_key_value = self.api_key.get_secret_value()

        # API keys are in format: moxn_{tenant_id}.{rest}
        # Extract the part after 'moxn_' and before the next dot
        if not api_key_value.startswith("moxn_"):
            raise ValueError("Invalid API key format - must start with 'moxn_'")

        # Remove 'moxn_' prefix and get tenant part
        key_without_prefix = api_key_value[5:]  # Remove 'moxn_'
        tenant_id = key_without_prefix.split(".")[0]

        if not tenant_id:
            raise ValueError("Invalid API key format - no tenant ID found")

        return tenant_id

    @staticmethod
    def _should_use_external_storage(content: dict) -> bool:
        """Check if content should be stored externally based on size"""
        try:
            return len(_content_adapter.dump_json(content)) > MAX_INLINE_CONTENT_SIZE
        except (TypeError, ValueError):
            return True

    async def _send_telemetry_log_with_external_content(
        self, log_request: TelemetryLogRequest
    ) -> TelemetryLogResponse:
        """Send telemetry log with content stored externally"""
        original_content = log_request.content
        log_request.content = None  # Clear content from inline request
        log_request.content_stored = True

        # Extract tenant ID from API key (format: moxn_{tenant_id}.{rest})
        tenant_id = self._extract_tenant_id_from_api_key()

        # Create storage path based on event IDs
        event_id = log_request.id
        file_path = f"{tenant_id}/{log_request.span_id or event_id}/{event_id}.json"

        signed_req = SignedURLRequest(
            file_path=file_path,
            media_type="application/json",
            log_request=log_request,
        )
        signed_resp = await self._send_telemetry_log_and_get_signed_url(signed_req)

        # Debug logging
        import logging

        logger = logging.getLogger(__name__)
        logger.debug(f"Received signed URL from API: {signed_resp.url}")

        # Validate signed URL before attempting upload
        if not signed_resp.url:
            raise ValueError("Signed URL response is missing URL field")

        # Support multiple URL schemes for generic hosting scenarios
        VALID_URL_SCHEMES = [
            "http://",
            "https://",  # Standard web protocols
            "s3://",  # AWS S3
            "gs://",  # Google Cloud Storage
            "azblob://",  # Azure Blob Storage
            "file://",  # Local filesystem (for testing)
        ]

        # Allow custom schemes via environment variable
        import os

        custom_schemes = os.getenv("MOXN_TELEMETRY_CUSTOM_URL_SCHEMES", "").split(",")
        VALID_URL_SCHEMES.extend(
            [s.strip() + "://" for s in custom_schemes if s.strip()]
        )

        if not isinstance(signed_resp.url, str):
            raise ValueError(
                f"Signed URL must be a string, got {type(signed_resp.url)}"
            )

        # Check if URL starts with any valid scheme
        if not any(signed_resp.url.startswith(scheme) for scheme in VALID_URL_SCHEMES):
            # Try to parse it as a URL to see if it has a valid structure
            from urllib.parse import urlparse

            parsed = urlparse(signed_resp.url)
            if not parsed.scheme or not parsed.netloc:
                raise ValueError(
                    f"Invalid signed URL format: {signed_resp.url}. "
                    f"Expected URL with one of these schemes: {', '.join(VALID_URL_SCHEMES)} "
                    "or set MOXN_TELEMETRY_CUSTOM_URL_SCHEMES environment variable."
                )

        # Set storage key in request
        log_request.content_storage_key = signed_resp.file_path

        # upload the content anonymously
        put_resp = await self._anon_client.put(
            signed_resp.url,
            content=_content_adapter.dump_json(original_content),
            headers={"Content-Type": "application/json"},
        )
        put_resp.raise_for_status()

        return TelemetryLogResponse(
            id=signed_resp.id or uuid4(),
            timestamp=datetime.now(timezone.utc),
        )

    async def aclose(self) -> None:
        """Properly close all underlying HTTPX clients."""
        await self._client.aclose()
        await self._anon_client.aclose()



from __future__ import annotations

from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Literal,
    Protocol,
    Sequence,
    TypeVar,
    runtime_checkable,
)
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field, model_validator

from moxn.types.content import Author, MessageRole
from moxn.types.request_config import CompletionConfig

if TYPE_CHECKING:
    from moxn.types.tool import SdkTool

T = TypeVar("T")
U = TypeVar("U")


class VersionRef(BaseModel):
    """Explicitly specify either branch or commit for version reference."""

    branch_name: str | None = None
    commit_id: str | None = None

    @model_validator(mode="after")
    def validate_xor(self):
        """Ensure exactly one of branch_name or commit_id is provided."""
        if bool(self.branch_name) == bool(self.commit_id):
            raise ValueError("Exactly one of branch_name or commit_id must be provided")
        return self


class GitTrackedEntity(BaseModel):
    """Base class for entities tracked in git-like system."""

    branch_id: UUID | None = Field(
        None, alias="branchId", description="Current branch name/ID"
    )
    commit_id: UUID | None = Field(
        None, alias="commitId", description="Current commit SHA (null = working state)"
    )

    model_config = ConfigDict(populate_by_name=True)

    @model_validator(mode="after")
    def validate_version_xor(self):
        """Ensure exactly one of branch_id or commit_id is provided."""
        # Temporarily disabled for backend compatibility
        # TODO: Re-enable once backend is fixed to send proper data
        # if bool(self.branch_id) == bool(self.commit_id):
        #     raise ValueError("Exactly one of branch_id or commit_id must be provided")
        return self


class Branch(BaseModel):
    """Named pointer to a commit."""

    id: UUID
    name: str
    head_commit_id: UUID | None = Field(None, alias="headCommitId")  # Commit SHA

    model_config = ConfigDict(populate_by_name=True)


class Commit(BaseModel):
    """Represents a snapshot of all entities at a point in time."""

    id: str  # Commit SHA hash
    message: str
    created_at: datetime | None = Field(None, alias="createdAt")

    model_config = ConfigDict(populate_by_name=True)


class MoxnSchemaMetadata(BaseModel):
    """Metadata embedded in JSON Schema x-moxn-metadata field."""

    schema_id: UUID
    schema_version_id: UUID | None = None
    prompt_id: UUID | None = None  # Optional - not all schemas belong to prompts
    prompt_version_id: UUID | None = None
    task_id: UUID
    branch_id: UUID | None = None  # Can be None in API responses
    commit_id: str | None = None

    model_config = {"populate_by_name": True}


@runtime_checkable
class RenderableModel(Protocol):
    moxn_schema_metadata: ClassVar[MoxnSchemaMetadata]

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: Any = None,
        exclude: Any = None,
        context: dict[str, Any] | None = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
        serialize_as_any: bool = False,
    ) -> Any: ...

    def render(self, **kwargs: Any) -> Any: ...


# BaseHeaders removed - legacy authentication pattern


class MessageBase(GitTrackedEntity, Generic[T]):
    id: UUID = Field(default_factory=uuid4)
    name: str
    description: str | None = None
    author: Author = Field(default=Author.HUMAN)  # Default to HUMAN
    role: MessageRole
    blocks: Sequence[Sequence[T]]
    task_id: UUID | None = Field(None, alias="taskId")


class BaseSchema(GitTrackedEntity):
    id: UUID
    name: str
    description: str | None = None
    exportedJSON: str


class BasePrompt(GitTrackedEntity, Generic[T, U]):
    id: UUID
    name: str
    description: str | None = None
    task_id: UUID = Field(..., alias="taskId")
    messages: Sequence[T]
    input_schema: U = Field(..., alias="inputSchema")
    message_order: list[UUID] | None = Field(None, alias="messageOrder")
    completion_config: CompletionConfig | None = Field(None, alias="completionConfig")
    tools: Sequence[SdkTool] | None = Field(
        None,
        description="Tools attached to this prompt for function calling or structured output. "
        "Tools with toolType='tool' are function calling tools. "
        "Tools with toolType='structured_output' define response format schemas.",
    )


class BaseTask(GitTrackedEntity, Generic[T]):
    id: UUID
    name: str
    description: str | None = None
    prompts: Sequence[T]
    definitions: dict[str, Any] = Field(default_factory=dict)
    branches: list[Branch] = Field(default_factory=list)
    last_commit: Commit | None = Field(None, alias="lastCommit")


# Response types for API operations
class BranchHeadResponse(BaseModel):
    """Response when resolving a branch to its head commit."""

    branch_id: str = Field(alias="branchId")  # Branch UUID
    branch_name: str = Field(alias="branchName")
    task_id: UUID = Field(alias="taskId")
    head_commit_id: str | None = Field(alias="headCommitId")  # Commit SHA
    parent_commit_id: str | None = Field(
        None, alias="parentCommitId"
    )  # Parent commit SHA
    effective_commit_id: str = Field(alias="effectiveCommitId")  # The commit ID to use
    has_uncommitted_changes: bool = Field(alias="hasUncommittedChanges")
    last_committed_at: datetime | None = Field(None, alias="lastCommittedAt")
    is_default: bool = Field(alias="isDefault")
    metadata: dict[str, Any] | None = None

    model_config = ConfigDict(populate_by_name=True)


class CommitInfoResponse(BaseModel):
    """Response with commit information."""

    commit: Commit
    branch_name: str | None = None

    model_config = ConfigDict(populate_by_name=True)



from typing import Sequence

from pydantic import ConfigDict, Field

from moxn.types.base import BasePrompt, BaseSchema, BaseTask, MessageBase
from moxn.types.blocks.content_block import ContentBlockModel


class MessageDTO(MessageBase[ContentBlockModel]):
    """Message DTO for API communication - blocks are raw dicts."""

    blocks: Sequence[Sequence[ContentBlockModel]] = Field(
        repr=False, default_factory=list
    )

    model_config = ConfigDict(populate_by_name=True)


class SchemaDTO(BaseSchema):
    model_config = ConfigDict(populate_by_name=True)


class PromptDTO(BasePrompt[MessageDTO, SchemaDTO]):
    """Prompt DTO for API communication."""

    messages: Sequence[MessageDTO] = Field(default_factory=list)
    input_schema: SchemaDTO = Field(..., alias="inputSchema")

    model_config = ConfigDict(populate_by_name=True)


class TaskDTO(BaseTask[PromptDTO]):
    """Task DTO for API communication."""

    prompts: Sequence[PromptDTO] = Field(default_factory=list)
    # definitions inherited from BaseTask

    model_config = ConfigDict(populate_by_name=True)



"""DTOs for Prompt Studio LLM invocation feature.

This module defines request/response types for the studio invocation endpoint,
enabling the backend to execute LLM invocations through the API gateway for
counterfactual testing (testing prompt changes against historical observations).
"""

from __future__ import annotations

from enum import Enum
from typing import Any, Literal, Sequence
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field, SecretStr

from moxn.types.content import Provider
from moxn.types.dto import MessageDTO, SchemaDTO
from moxn.types.request_config import CompletionConfig, RequestConfig, SchemaDefinition
from moxn.types.telemetry import ResponseType
from moxn.types.tool import SdkTool


class LLMErrorCode(str, Enum):
    """Error codes for LLM invocation failures."""

    INVALID_API_KEY = "INVALID_API_KEY"
    QUOTA_EXCEEDED = "QUOTA_EXCEEDED"
    RATE_LIMITED = "RATE_LIMITED"
    PROVIDER_UNAVAILABLE = "PROVIDER_UNAVAILABLE"
    PROVIDER_TIMEOUT = "PROVIDER_TIMEOUT"
    INVALID_MODEL = "INVALID_MODEL"
    CONTENT_FILTERED = "CONTENT_FILTERED"
    CONTEXT_LENGTH_EXCEEDED = "CONTEXT_LENGTH_EXCEEDED"
    VARIABLE_MISSING = "VARIABLE_MISSING"
    COMPILATION_ERROR = "COMPILATION_ERROR"
    UNEXPECTED_ERROR = "UNEXPECTED_ERROR"


# =============================================================================
# Request Types
# =============================================================================


class StudioObservation(BaseModel):
    """Single observation for LLM invocation.

    An observation represents a historical invocation with its messages and
    rendered input values. Messages may include conversation context (previous
    LLM responses, tool calls) merged with template messages.
    """

    id: UUID = Field(
        default_factory=uuid4, description="Unique ID for this observation"
    )
    rendered_input: dict[str, Any] = Field(
        ...,
        alias="renderedInput",
        description="Flat key-value pairs matching prompt variable names",
    )
    messages: Sequence[MessageDTO] = Field(
        ...,
        description="Messages for this observation (with variable blocks to hydrate)",
    )

    model_config = ConfigDict(populate_by_name=True)


class StudioExecutionMetadata(BaseModel):
    """Tracking metadata from backend - echoed in response.

    The backend can include any fields needed for tracking; they are
    echoed back in the response without modification.
    """

    execution_id: UUID = Field(..., alias="executionId")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class StudioPromptDTO(BaseModel):
    """Prompt DTO for studio invocations - messages are per-observation.

    This type contains prompt metadata and configuration but excludes messages,
    which are now provided per-observation to support multi-turn conversations.
    """

    id: UUID
    name: str
    description: str | None = None
    task_id: UUID = Field(..., alias="taskId")
    input_schema: SchemaDTO = Field(..., alias="inputSchema")
    completion_config: CompletionConfig | None = Field(None, alias="completionConfig")
    tools: Sequence[SdkTool] | None = Field(
        None,
        description="Tools attached to this prompt for function calling or structured output.",
    )

    model_config = ConfigDict(populate_by_name=True)


class StudioInvocationRequest(BaseModel):
    """Request to execute LLM invocations for Prompt Studio.

    Enables counterfactual testing: execute a (potentially modified) prompt
    against historical observations to see how the output would change.

    Messages are now per-observation to support multi-turn conversations where
    each observation may have different conversation context.
    """

    # Prompt definition (without messages - messages are per-observation)
    prompt: StudioPromptDTO = Field(..., description="Prompt metadata and configuration")

    # Version context
    task_id: UUID = Field(..., alias="taskId")
    branch_id: UUID | None = Field(None, alias="branchId")
    commit_id: str | None = Field(None, alias="commitId")

    # Observations to process (1-10)
    observations: Sequence[StudioObservation] = Field(
        ...,
        min_length=1,
        max_length=10,
        description="1-10 observations to evaluate against the prompt",
    )

    # Provider configuration
    completion_config: CompletionConfig = Field(..., alias="completionConfig")
    provider_api_key: SecretStr = Field(
        ...,
        alias="providerApiKey",
        description="Provider API key (passed per-request, not stored)",
    )

    # Tracking metadata (echoed back in response)
    execution_metadata: StudioExecutionMetadata = Field(..., alias="executionMetadata")

    model_config = ConfigDict(populate_by_name=True)


# =============================================================================
# Response Types - Aligned with LLMEventModelBase
# =============================================================================


class StudioObservationResult(BaseModel):
    """Result for a single observation - mirrors LLMEventModelBase fields.

    Contains both the LLM response data (aligned with telemetry events)
    and studio-specific fields like latency and error details.
    """

    # Observation identification
    observation_id: UUID = Field(..., alias="observationId")

    # LLMEventModelBase-aligned fields
    prompt_id: UUID = Field(..., alias="promptId")
    prompt_name: str = Field(..., alias="promptName")
    branch_id: UUID | None = Field(None, alias="branchId")
    commit_id: str | None = Field(None, alias="commitId")
    messages: list[dict[str, Any]] = Field(
        default_factory=list,
        description="Messages used in the invocation (serialized with blocks)",
    )
    provider: Provider
    raw_response: dict[str, Any] | None = Field(
        None,
        alias="rawResponse",
        description="Raw response from the LLM provider",
    )
    parsed_response: dict[str, Any] | None = Field(
        None,
        alias="parsedResponse",
        description="Normalized response with content blocks (pre-serialized dict)",
    )
    rendered_input: dict[str, Any] = Field(
        ...,
        alias="renderedInput",
        description="The observation's rendered input (echoed back)",
    )
    response_type: ResponseType | None = Field(
        None,
        alias="responseType",
        description="Classification of response type (text, tool_calls, etc.)",
    )
    request_config: RequestConfig | None = Field(
        None,
        alias="requestConfig",
        description="Provider-specific request configuration used",
    )
    schema_definition: SchemaDefinition | None = Field(
        None,
        alias="schemaDefinition",
        description="Schema or tool definitions used in the request",
    )
    validation_errors: list[str] | None = Field(
        None,
        alias="validationErrors",
        description="Schema validation errors if any occurred",
    )
    tool_calls_count: int = Field(
        0,
        alias="toolCallsCount",
        description="Number of tool calls in the response",
    )
    attributes: dict[str, Any] | None = Field(
        None,
        description="Observation attributes (provider, responseType, etc.)",
    )

    # Studio-specific fields
    status: Literal["success", "error"] = Field(
        ...,
        description="Whether the invocation succeeded or failed",
    )
    latency_ms: float = Field(
        ...,
        alias="latencyMs",
        description="Execution time in milliseconds",
    )
    error_code: LLMErrorCode | None = Field(
        None,
        alias="errorCode",
        description="Error code if status is 'error'",
    )
    error_message: str | None = Field(
        None,
        alias="errorMessage",
        description="Human-readable error message if status is 'error'",
    )

    model_config = ConfigDict(populate_by_name=True)


class StudioInvocationResponse(BaseModel):
    """Response from studio invocation endpoint.

    Contains results for all observations plus aggregate statistics.
    Partial failures are supported - some observations may succeed while others fail.
    """

    # Echoed tracking metadata
    execution_metadata: StudioExecutionMetadata = Field(..., alias="executionMetadata")

    # Per-observation results
    results: Sequence[StudioObservationResult]

    # Aggregate statistics
    total_observations: int = Field(..., alias="totalObservations")
    successful_count: int = Field(..., alias="successfulCount")
    failed_count: int = Field(..., alias="failedCount")

    model_config = ConfigDict(populate_by_name=True)



from typing import TYPE_CHECKING, Any, Sequence

from pydantic import Field

from moxn.models import message as msg
from moxn.models import schema as sch
from moxn.types.base import BasePrompt, RenderableModel
from moxn.types.content import MessageRole
from moxn.types.tool import SdkSchema, SdkTool

if TYPE_CHECKING:
    PromptSession = Any


class PromptTemplate(BasePrompt[msg.Message, sch.Schema]):
    """Immutable representation of a stored prompt configuration.

    Prompts can have attached tools for function calling and/or structured output.
    Tools are stored in the `tools` field (inherited from BasePrompt) and can be:
    - Function calling tools (tool_type='tool')
    - Structured output schemas (tool_type='structured_output')
    """

    messages: Sequence[msg.Message]
    input_schema: sch.Schema = Field(..., alias="inputSchema")

    @property
    def function_tools(self) -> list[SdkTool]:
        """Get tools for function calling (tool_type='tool' only).

        Returns:
            List of SdkTool objects that are function calling tools,
            sorted by position.
        """
        if not self.tools:
            return []
        return sorted(
            [t for t in self.tools if t.tool_type == "tool"],
            key=lambda t: t.position,
        )

    @property
    def structured_output_schema(self) -> SdkSchema | None:
        """Get the structured output schema if response_schema_id is set.

        Looks up the schema in the tools array by ID - no fetch, all data is local.
        The schema is only returned if:
        1. completion_config.response_schema_id is set
        2. A matching tool with tool_type='structured_output' exists

        Returns:
            SdkSchema for structured output, or None if not configured.
        """
        if not self.completion_config or not self.completion_config.response_schema_id:
            return None

        if not self.tools:
            return None

        response_schema_id = self.completion_config.response_schema_id
        for tool in self.tools:
            if (
                tool.tool_type == "structured_output"
                and str(tool.schema_.id) == response_schema_id
            ):
                return tool.schema_

        return None

    def get_tool_by_name(self, name: str) -> SdkTool | None:
        """Lookup a tool by name.

        Args:
            name: The name of the tool (matches schema.name)

        Returns:
            SdkTool if found, None otherwise.
        """
        if not self.tools:
            return None

        for tool in self.tools:
            if tool.schema_.name == name:
                return tool

        return None

    def get_messages(self):
        return [message.model_copy(deep=True) for message in self.messages]

    def get_message_by_role(self, role: str | MessageRole) -> msg.Message | None:
        """Get the first message with the specified role."""
        _role = MessageRole(role) if isinstance(role, str) else role

        messages = [p for p in self.messages if p.role == role]
        if len(messages) == 1:
            return messages[0]
        elif len(messages) == 0:
            return None
        else:
            raise ValueError(
                f"get message is not deterministic, there are {len(messages)} {_role.value} messages in the prompt"
            )

    def get_messages_by_role(self, role: str | MessageRole) -> list[msg.Message]:
        """Get all messages with the specified role."""
        role = MessageRole(role) if isinstance(role, str) else role
        return [p for p in self.messages if p.role == role]

    def to_prompt_session(
        self,
        session_data: RenderableModel | None = None,
        render_kwargs: dict[str, Any] | None = None,
    ) -> "PromptSession":  # type: ignore
        from moxn.models.prompt.session import PromptSession

        return PromptSession.from_prompt_template(
            prompt=self,
            session_data=session_data,
            render_kwargs=render_kwargs,
        )



"""Raw content block model for flexible API parsing."""

from typing import Any
from pydantic import BaseModel, ConfigDict, Field


class RawContentBlock(BaseModel):
    """
    Flexible content block model for parsing API responses.

    This model is designed to handle any block type from the API without
    breaking on unknown fields or new block types. It captures the common
    structure while allowing extra fields.
    """

    # Core fields present in all blocks
    block_type: str = Field(alias="blockType")
    options: dict[str, Any] = Field(default_factory=dict)

    # Text block fields
    text: str | None = None

    # Variable block fields
    name: str | None = None
    variable_type: str | None = Field(None, alias="variableType")
    format: str | None = None
    description: str | None = None
    required: bool | None = None
    default_value: str | None = Field(None, alias="defaultValue")

    # Image/File source type fields
    type: str | None = None  # For discriminating image/file sources

    # Image/File content fields - these vary by source type
    base64: str | None = None
    url: str | None = None
    uri: str | None = None
    filepath: str | None = None
    bytes_data: bytes | None = None  # Renamed to avoid conflict with built-in bytes

    # Media type for images/files
    media_type: str | None = Field(None, alias="mediaType")

    # Signed URL fields
    file_path: str | None = Field(None, alias="filePath")

    # Tool result/call fields (for future extension)
    tool_name: str | None = Field(None, alias="toolName")
    tool_call_id: str | None = Field(None, alias="toolCallId")

    model_config = ConfigDict(
        populate_by_name=True,
        extra="allow",  # Allow extra fields from API
    )

    def get_extra_fields(self) -> dict[str, Any]:
        """Get any extra fields not defined in the model."""
        return {k: v for k, v in self.__dict__.items() if k not in self.model_fields}



"""Request models for write operations (creates) shared between SDK and API."""

from typing import Sequence
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from moxn.types.blocks.content_block import ContentBlockModel
from moxn.types.content import Author, MessageRole


class TaskCreateRequest(BaseModel):
    """Request to create a new task."""

    name: str = Field(..., min_length=1, max_length=255, description="Task name")
    description: str | None = Field(
        None, max_length=1000, description="Optional task description"
    )
    branch_name: str = Field(
        "main", alias="branchName", description="Branch to create task on"
    )

    model_config = ConfigDict(populate_by_name=True)


class MessageData(BaseModel):
    """Message data for inline creation within prompts.

    Messages are created as part of prompt creation, not as standalone entities.
    """

    name: str = Field(..., description="Message name")
    role: MessageRole = Field(
        ..., description="Message role (system, user, assistant, etc.)"
    )
    author: Author = Field(
        Author.HUMAN, description="Message author (HUMAN or MACHINE)"
    )
    description: str | None = Field(None, description="Optional message description")
    blocks: Sequence[Sequence[ContentBlockModel]] = Field(
        ..., description="2D array of content blocks (paragraphs containing blocks)"
    )

    model_config = ConfigDict(populate_by_name=True)


class PromptCreateRequest(BaseModel):
    """Request to create a prompt with messages.

    The backend will:
    - Create the prompt
    - Create all messages inline
    - Auto-generate the input schema from variables in messages
    - Handle property deduplication for variables with same name
    """

    name: str = Field(..., min_length=1, max_length=255, description="Prompt name")
    task_id: UUID = Field(..., alias="taskId", description="Parent task ID")
    description: str | None = Field(None, description="Optional prompt description")
    branch_name: str = Field(
        "main", alias="branchName", description="Branch to create prompt on"
    )
    messages: Sequence[MessageData] = Field(
        ..., description="Messages to create with the prompt (at least one required)"
    )

    model_config = ConfigDict(populate_by_name=True)



"""Type definitions for OpenAI Responses API provider-specific content blocks."""

from typing import TYPE_CHECKING, Sequence, TypedDict, Union

if TYPE_CHECKING:
    # Import from openai SDK - Responses API types
    from openai.types.responses import Response as OpenAIResponse
    from openai.types.responses import (
        ResponseFunctionToolCallParam as OpenAIResponsesFunctionToolCallParam,
    )
    from openai.types.responses import (
        ResponseInputContentParam as OpenAIResponsesInputContent,
    )
    from openai.types.responses import (
        ResponseInputFileParam as OpenAIResponsesInputFileParam,
    )
    from openai.types.responses import (
        ResponseInputImageParam as OpenAIResponsesInputImageParam,
    )
    from openai.types.responses import (
        ResponseInputItemParam as OpenAIResponsesInputItemParam,
    )
    from openai.types.responses import ResponseInputParam as OpenAIResponsesInputParam
    from openai.types.responses import (
        ResponseInputTextParam as OpenAIResponsesInputTextParam,
    )
    from openai.types.responses import ResponseOutputItem as OpenAIResponsesOutputItem
    from openai.types.responses import (
        ResponseOutputMessage as OpenAIResponsesOutputMessage,
    )
    from openai.types.responses.response_input_item_param import (
        FunctionCallOutput as OpenAIResponsesFunctionCallOutput,
    )
    from openai.types.responses.response_input_item_param import (
        Message as OpenAIResponsesMessageItemParam,
    )
else:
    # Runtime fallbacks when OpenAI SDK not installed
    OpenAIResponse = dict
    OpenAIResponsesInputContent = dict
    OpenAIResponsesInputItemParam = dict
    OpenAIResponsesInputParam = list
    OpenAIResponsesInputTextParam = dict
    OpenAIResponsesInputImageParam = dict
    OpenAIResponsesInputFileParam = dict
    OpenAIResponsesFunctionToolCallParam = dict
    OpenAIResponsesFunctionCallOutput = dict
    OpenAIResponsesMessageItemParam = dict
    OpenAIResponsesInputItemParam = dict
    OpenAIResponsesOutputItem = dict
    OpenAIResponsesOutputMessage = dict
    try:
        from openai.types.responses import Response as OpenAIResponse
    except ImportError:
        pass

# Content block types (within Message items and as separate items)
OpenAIResponsesContentBlock = Union[
    OpenAIResponsesInputTextParam,
    OpenAIResponsesInputImageParam,
    OpenAIResponsesInputFileParam,
    OpenAIResponsesFunctionToolCallParam,
    OpenAIResponsesFunctionCallOutput,
]

# Provider-specific block sequences (for document operations)
OpenAIResponsesContentBlockSequence = Sequence[Sequence[OpenAIResponsesContentBlock]]

# Item-level sequences (for items that can be in input/output arrays)
OpenAIResponsesInputItemSequence = Sequence[OpenAIResponsesInputItemParam]


class OpenAIResponsesMessagesParam(TypedDict, total=False):
    """Payload structure for OpenAI Responses API.

    This matches the structure of ResponseCreateParams from the OpenAI SDK,
    but only includes the fields we generate (input and instructions).
    """

    input: list[OpenAIResponsesInputItemParam]
    """List of input items (messages, tool calls, tool results, etc.)."""

    instructions: str
    """System/developer instructions extracted from system-role messages."""



"""Tool formatting utilities for converting Moxn tools to provider-specific formats.

This module provides formatters for:
1. ToolFormatter: Converts SdkTool definitions to provider-specific tool formats
2. ToolChoiceTranslator: Translates Moxn tool_choice to provider-specific formats

Provider Translation Reference:

Tool Choice:
| Moxn Value      | OpenAI                                        | Anthropic              | Google                                           |
|-----------------|-----------------------------------------------|------------------------|--------------------------------------------------|
| 'auto'          | 'auto'                                        | {type: 'auto'}         | {mode: 'AUTO'}                                   |
| 'none'          | 'none'                                        | (omit tools)           | {mode: 'NONE'}                                   |
| 'required'      | 'required'                                    | {type: 'any'}          | {mode: 'ANY'}                                    |
| {tool: 'name'}  | {type: 'function', function: {name: 'name'}}  | {type: 'tool', name}   | {mode: 'ANY', allowed_function_names: ['name']}  |

Tool Definition:
| Provider   | Structure                                                           |
|------------|---------------------------------------------------------------------|
| OpenAI     | {type: 'function', function: {name, description, parameters}}       |
| Anthropic  | {name, description, input_schema}                                   |
| Google     | {function_declarations: [{name, description, parameters}]}          |
"""

from __future__ import annotations

import json
from typing import Any

from moxn.types.tool import SdkTool


class ToolFormatter:
    """Converts SdkTool definitions to provider-specific tool formats."""

    @staticmethod
    def to_openai_tools(tools: list[SdkTool]) -> list[dict[str, Any]]:
        """Convert SdkTool list to OpenAI tool format.

        OpenAI format:
        {
            "type": "function",
            "function": {
                "name": "tool_name",
                "description": "Tool description",
                "parameters": { ... JSON Schema ... }
            }
        }

        Args:
            tools: List of SdkTool objects (should be filtered to tool_type='tool')

        Returns:
            List of OpenAI tool definitions
        """
        result = []
        for tool in tools:
            schema = tool.schema_
            parameters = json.loads(schema.exported_json)

            result.append(
                {
                    "type": "function",
                    "function": {
                        "name": schema.name,
                        "description": schema.description or "",
                        "parameters": parameters,
                    },
                }
            )
        return result

    @staticmethod
    def to_anthropic_tools(tools: list[SdkTool]) -> list[dict[str, Any]]:
        """Convert SdkTool list to Anthropic tool format.

        Anthropic format:
        {
            "name": "tool_name",
            "description": "Tool description",
            "input_schema": { ... JSON Schema ... }
        }

        Args:
            tools: List of SdkTool objects (should be filtered to tool_type='tool')

        Returns:
            List of Anthropic tool definitions
        """
        result = []
        for tool in tools:
            schema = tool.schema_
            input_schema = json.loads(schema.exported_json)

            result.append(
                {
                    "name": schema.name,
                    "description": schema.description or "",
                    "input_schema": input_schema,
                }
            )
        return result

    @staticmethod
    def to_google_function_declarations(tools: list[SdkTool]) -> list[dict[str, Any]]:
        """Convert SdkTool list to Google function declarations.

        Google format (function declarations, not wrapped in tools):
        {
            "name": "tool_name",
            "description": "Tool description",
            "parameters": { ... JSON Schema ... }
        }

        Note: Google wraps these in a Tool object with function_declarations array.
        This method returns the declarations; the caller wraps them.

        Args:
            tools: List of SdkTool objects (should be filtered to tool_type='tool')

        Returns:
            List of Google function declarations
        """
        result = []
        for tool in tools:
            schema = tool.schema_
            parameters = json.loads(schema.exported_json)

            result.append(
                {
                    "name": schema.name,
                    "description": schema.description or "",
                    "parameters": parameters,
                }
            )
        return result


class ToolChoiceTranslator:
    """Translates Moxn tool_choice to provider-specific formats."""

    @staticmethod
    def to_openai(choice: str | dict[str, Any]) -> str | dict[str, Any]:
        """Translate Moxn tool_choice to OpenAI format.

        Translations:
        - 'auto' -> 'auto'
        - 'none' -> 'none'
        - 'required' -> 'required'
        - {tool: 'name'} -> {type: 'function', function: {name: 'name'}}

        Args:
            choice: Moxn tool_choice value

        Returns:
            OpenAI tool_choice value
        """
        if isinstance(choice, str):
            # 'auto', 'none', 'required' pass through unchanged
            return choice

        if isinstance(choice, dict) and "tool" in choice:
            # {tool: 'name'} -> {type: 'function', function: {name: 'name'}}
            return {
                "type": "function",
                "function": {"name": choice["tool"]},
            }

        # Pass through any other dict format unchanged
        return choice

    @staticmethod
    def to_anthropic(choice: str | dict[str, Any]) -> dict[str, Any] | None:
        """Translate Moxn tool_choice to Anthropic format.

        Translations:
        - 'auto' -> {type: 'auto'}
        - 'none' -> None (caller should omit tools)
        - 'required' -> {type: 'any'}
        - {tool: 'name'} -> {type: 'tool', name: 'name'}

        Args:
            choice: Moxn tool_choice value

        Returns:
            Anthropic tool_choice value, or None if tools should be omitted
        """
        if isinstance(choice, str):
            if choice == "auto":
                return {"type": "auto"}
            elif choice == "none":
                # Return None to signal that tools should be omitted entirely
                return None
            elif choice == "required":
                return {"type": "any"}
            else:
                # Unknown string, pass through as type
                return {"type": choice}

        if isinstance(choice, dict) and "tool" in choice:
            # {tool: 'name'} -> {type: 'tool', name: 'name'}
            return {"type": "tool", "name": choice["tool"]}

        # Pass through any other dict format unchanged
        return choice

    @staticmethod
    def to_google(choice: str | dict[str, Any]) -> dict[str, Any]:
        """Translate Moxn tool_choice to Google format.

        Translations:
        - 'auto' -> {mode: 'AUTO'}
        - 'none' -> {mode: 'NONE'}
        - 'required' -> {mode: 'ANY'}
        - {tool: 'name'} -> {mode: 'ANY', allowed_function_names: ['name']}

        Args:
            choice: Moxn tool_choice value

        Returns:
            Google function_calling_config value
        """
        if isinstance(choice, str):
            if choice == "auto":
                return {"mode": "AUTO"}
            elif choice == "none":
                return {"mode": "NONE"}
            elif choice == "required":
                return {"mode": "ANY"}
            else:
                # Unknown string, try to use as mode
                return {"mode": choice.upper()}

        if isinstance(choice, dict) and "tool" in choice:
            # {tool: 'name'} -> {mode: 'ANY', allowed_function_names: ['name']}
            return {
                "mode": "ANY",
                "allowed_function_names": [choice["tool"]],
            }

        # Pass through any other dict format unchanged
        return choice



"""Combined invocation TypedDicts for direct SDK unpacking.

These TypedDicts combine message payloads with model parameters,
enabling direct unpacking into provider SDK calls:

    response = await anthropic.messages.create(**session.to_anthropic_invocation())
"""

from typing import Sequence, TypedDict

from moxn.types.type_aliases.anthropic import (
    AnthropicContentBlockParam,
    AnthropicSystemContentBlockParam,
)
from moxn.types.type_aliases.google import GoogleContent, GoogleFile
from moxn.types.type_aliases.openai_chat import OpenAIChatContentBlock
from moxn.types.type_aliases.openai_responses import OpenAIResponsesInputItemParam


class AnthropicInvocationParam(TypedDict, total=False):
    """Complete Anthropic payload for `anthropic.messages.create(**payload)`.

    Combines message content with model parameters for direct SDK unpacking.
    """

    # Message fields (from AnthropicMessagesParam)
    system: str | list[AnthropicSystemContentBlockParam]
    messages: list[AnthropicContentBlockParam]

    # Model parameters
    model: str
    max_tokens: int
    temperature: float
    top_p: float

    # Thinking/reasoning parameters
    thinking: dict


class OpenAIChatInvocationParam(TypedDict, total=False):
    """Complete OpenAI Chat payload for `openai.chat.completions.create(**payload)`.

    Combines message content with model parameters for direct SDK unpacking.
    """

    # Message fields (from OpenAIChatMessagesParam)
    messages: list[OpenAIChatContentBlock]

    # Model parameters
    model: str
    max_tokens: int
    temperature: float
    top_p: float

    # Thinking/reasoning parameters
    reasoning_effort: str


class OpenAIResponsesInvocationParam(TypedDict, total=False):
    """Complete OpenAI Responses API payload for `openai.responses.create(**payload)`.

    Combines input items with model parameters for direct SDK unpacking.
    """

    # Message fields (from OpenAIResponsesMessagesParam)
    input: list[OpenAIResponsesInputItemParam]
    instructions: str

    # Model parameters
    model: str
    max_tokens: int
    temperature: float
    top_p: float

    # Thinking/reasoning parameters
    reasoning: dict


class GoogleGenerateContentConfigDict(TypedDict, total=False):
    """Config options for Google genai `generate_content()`.

    These parameters are nested under the `config` key when calling
    `client.models.generate_content(model=..., contents=..., config=...)`.
    """

    # System instruction
    system_instruction: str

    # Generation parameters
    temperature: float
    top_p: float
    top_k: int
    max_output_tokens: int
    stop_sequences: list[str]
    presence_penalty: float
    frequency_penalty: float

    # Tool configuration
    tools: list[dict]
    tool_config: dict

    # Structured output
    response_schema: dict
    response_mime_type: str

    # Thinking/reasoning (for models that support it)
    thinking_config: dict


class GoogleInvocationParam(TypedDict, total=False):
    """Complete Google Gemini/Vertex payload for `generate_content(**payload)`.

    Designed for direct SDK unpacking:
        payload = session.to_google_gemini_invocation()
        response = client.models.generate_content(**payload)

    Note: Google genai SDK uses:
    - `contents` (plural) not `content`
    - Nested `config` for all generation parameters
    """

    # Top-level parameters
    model: str
    contents: Sequence[GoogleContent | GoogleFile]

    # All config params nested here
    config: GoogleGenerateContentConfigDict


# Union type for any provider invocation payload
ProviderInvocationPayload = (
    AnthropicInvocationParam
    | OpenAIChatInvocationParam
    | OpenAIResponsesInvocationParam
    | GoogleInvocationParam
)



from enum import Enum
from typing import Annotated, Literal

from pydantic import ConfigDict, Field

from moxn.types.blocks.base import BaseContent, BlockType


# 1. Enums
# --------------------------------------------------------------------------- #
class VariableFormat(str, Enum):
    INLINE = "inline"
    BLOCK = "block"


class VariableType(str, Enum):
    """Format of a variable."""

    PRIMITIVE = "primitive"
    IMAGE = "image"
    FILE = "file"


# --------------------------------------------------------------------------- #
# 2. Base Variable class
# --------------------------------------------------------------------------- #


class TextVariableModel(BaseContent):
    """A variable that represents text content."""

    block_type: Literal[BlockType.VARIABLE] = Field(
        default=BlockType.VARIABLE, alias="blockType"
    )
    name: str
    variable_type: Literal[VariableType.PRIMITIVE] = Field(
        default=VariableType.PRIMITIVE, alias="variableType"
    )
    format: VariableFormat
    description: str = ""
    required: bool = True
    default_value: str | None = Field(None, alias="defaultValue")

    model_config = ConfigDict(populate_by_name=True)


class ImageVariableModel(BaseContent):
    """A variable that represents image content."""

    block_type: Literal[BlockType.VARIABLE] = Field(
        default=BlockType.VARIABLE, alias="blockType"
    )
    name: str
    variable_type: Literal[VariableType.IMAGE] = Field(
        default=VariableType.IMAGE, alias="variableType"
    )
    format: VariableFormat
    description: str = ""
    required: bool = True

    model_config = ConfigDict(populate_by_name=True)


class FileVariableModel(BaseContent):
    """A variable that represents document content (PDF)."""

    block_type: Literal[BlockType.VARIABLE] = Field(
        default=BlockType.VARIABLE, alias="blockType"
    )
    name: str
    variable_type: Literal[VariableType.FILE] = Field(
        default=VariableType.FILE, alias="variableType"
    )
    format: VariableFormat
    description: str = ""
    required: bool = True

    model_config = ConfigDict(populate_by_name=True)


VariableContentModel = Annotated[
    TextVariableModel | ImageVariableModel | FileVariableModel,
    Field(discriminator="variable_type"),
]
